@@ -32,7 +32,6 @@
   #:use-module (shepherd support)
   #:use-module (shepherd service)
   #:use-module (shepherd system)
-  #:use-module (shepherd runlevel)
   #:use-module (shepherd args)
   #:use-module (shepherd comm)
   #:export (main))
@@ -216,12 +215,19 @@
         (sigaction SIGALRM (lambda _ (alarm 1)))
         (alarm 1))
 
-      ;; Stop everything when we get SIGINT.  When running as PID 1, that means
-      ;; rebooting; this is what happens when pressing ctrl-alt-del, see
-      ;; ctrlaltdel(8).
+      (when (= 1 (getpid))
+        ;; When running as PID 1, disable hard reboots upon ctrl-alt-del.
+        ;; Instead, the kernel will send us SIGINT so that we can gracefully
+        ;; shut down.  See ctrlaltdel(8) and kernel/reboot.c.
+        (disable-reboot-on-ctrl-alt-del))
+
+      ;; Stop everything when we get SIGINT.
       (sigaction SIGINT
         (lambda _
-          (stop root-service)))
+          (catch 'quit
+            (lambda ()
+              (stop root-service))
+            quit-exception-handler)))
 
       ;; Stop everything when we get SIGTERM.
       (sigaction SIGTERM
@@ -241,12 +247,7 @@
           ;; Get commands from the standard input port.
           (process-textual-commands (current-input-port))
           ;; Process the data arriving at a socket.
-          (let ((sock   (open-server-socket socket-file))
-
-                ;; With Guile <= 2.0.9, we can get a system-error exception for
-                ;; EINTR, which happens anytime we receive a signal, such as
-                ;; SIGCHLD.  Thus, wrap the 'accept' call.
-                (accept (EINTR-safe accept)))
+          (let ((sock   (open-server-socket socket-file)))
 
             ;; Possibly write out our PID, which means we're ready to accept
             ;; connections.  XXX: What if we daemonized already?
@@ -261,7 +262,7 @@
               (define (read-from sock)
                 (match (accept sock)
                   ((command-source . client-address)
-                   (setvbuf command-source _IOFBF 1024)
+                   (setvbuf command-source (buffering-mode block) 1024)
                    (process-connection command-source))
                   (_ #f)))
               (match (select (list sock) (list) (list) (if poll-services? 0.5 #f))
@@ -273,6 +274,26 @@
                 (check-for-dead-services))
               (next-command)))))))
 
+;; Start all of SERVICES, which is a list of canonical names (FIXME?),
+;; but in a order where all dependencies are fulfilled before we
+;; attempt to start a service.  It is assumed that this is possible (FIXME?).
+;; FIXME: This procedure is untested!
+(define (start-in-order services)
+  ;; Same thing with an association list where the cdr of each pair is
+  ;; `#f' if the service is not yet started.
+  (define (start-in-order-assoc assoc-services)
+    (while (any not (map cdr assoc-services))
+      (for-each (lambda (x)
+		  (let ((service (car (lookup-services (car x)))))
+		    (and (not (cdr x))
+			 (depends-resolved? service)
+			 (set-cdr! x (start service)))))
+		assoc-services)))
+
+  (start-in-order-assoc (map (lambda (x)
+			       (cons x #f))
+			     services)))
+
 (define (process-connection sock)
   "Process client connection SOCK, reading and processing commands."
   (catch 'system-error
@@ -289,8 +310,15 @@
       ;; Maybe we got EPIPE while writing to SOCK, or something like that.
       (false-if-exception (close sock)))))
 
-(define %not-newline
-  (char-set-complement (char-set #\newline)))
+(define (quit-exception-handler key)
+  "Handle the 'quit' exception, rebooting if we're running as root."
+  ;; Most likely we're receiving 'quit' from the 'stop' method of
+  ;; ROOT-SERVICE.  So, if we're running as 'root', just reboot.
+  (if (zero? (getuid))
+      (begin
+        (local-output (l10n "Rebooting..."))
+        (reboot))
+      (quit)))
 
 (define (process-command command port)
   "Interpret COMMAND, a command sent by the user, represented as a
@@ -307,8 +335,19 @@
              (open-output-string)))
 
          (define (get-messages)
-           (string-tokenize (get-output-string message-port)
-                            %not-newline))
+           (let* ((str (get-output-string message-port))
+                  (lst (string-split str #\newline)))
+             ;; 'string-tokenize' swallows empty lines, which is not great,
+             ;; and 'string-split' doesn't distinguish between an empty line
+             ;; and this empty string, which is not great either.  So we hack
+             ;; our way the best we can.
+             (cond ((string-null? str)
+                    '())
+                   ;; If STR ends in \n, drop the trailing empty string since
+                   ;; that would lead the client to print an extra newline.
+                   ((string-suffix? "\n" str)
+                    (drop-right lst 1))
+                   (else lst))))
 
          (parameterize ((%current-client-socket message-port))
            (guard (c ((service-error? c)
@@ -330,16 +369,9 @@
 
              (write-reply (command-reply command result #f (get-messages))
                           port))))
-       (lambda (key)
-         ;; Most likely we're receiving 'quit' from the 'stop' method of
-         ;; ROOT-SERVICE.  So, if we're running as 'root', just reboot.
-         (if (zero? (getuid))
-             (begin
-               (local-output "Rebooting...")
-               (reboot))
-             (quit)))))
+       quit-exception-handler))
     (_
-     (local-output "Invalid command."))))
+     (local-output (l10n "Invalid command.")))))
 
 (define (process-textual-commands port)
   "Process textual commands from PORT.  'Textual' means that they're as you
@@ -358,5 +390,5 @@ would write them on the 'herd' command line."
                                                 #:arguments arguments)
                               port))
             (_
-             (local-output "invalid command line" line)))
+             (local-output (l10n "invalid command line") line)))
           (loop (read-line port))))))
