@@ -83,7 +83,7 @@ return the socket."
       (catch 'system-error
         (lambda ()
           (connect sock address)
-          (setvbuf sock _IOFBF 1024))
+          (setvbuf sock (buffering-mode block) 1024))
         (lambda (key proc format-string format-args errno . rest)
           ;; Guile's 'connect' throws an exception that doesn't specify
           ;; FILE.  Augment it with this information.
@@ -240,7 +240,7 @@ mechanism."
             (catch 'system-error
               (lambda ()
                 (connect sock AF_UNIX "/dev/log")
-                (setvbuf sock _IOLBF)
+                (setvbuf sock (buffering-mode line))
                 (set! port sock)
                 (call/syslog))
               (lambda args
@@ -251,14 +251,14 @@ mechanism."
                       (lambda ()
                         (call-with-output-file "/dev/kmsg"
                           (lambda (port)
-                            (setvbuf port _IOFBF)
+                            (setvbuf port (buffering-mode block))
                             (proc port))))
                       (lambda args
                         (if (memv (system-error-errno args)
                                   (list ENOENT EACCES EPERM))
                             (call-with-output-file "/dev/console"
                               (lambda (port)
-                                (setvbuf port _IONBF)
+                                (setvbuf port (buffering-mode none))
                                 (proc port)))
                             (apply throw args))))
                     (apply throw args)))))))))
@@ -282,9 +282,6 @@ available."
       (call-with-syslog-port close-port)))
    "w"))                                          ;output port
 
-(define %not-newline
-  (char-set-complement (char-set #\newline)))
-
 ;; We provide our own output mechanism, because we have certain
 ;; special needs; most importantly, we want to send output to herd
 ;; sometimes.
@@ -311,18 +308,23 @@ available."
         ;; completed line.
         (if (not (string-index str #\newline))
             (set! buffer (cons str buffer))
-            (let* ((str   (string-concatenate-reverse (cons str buffer)))
-                   (lines (string-tokenize str %not-newline)))
+            (let ((str (string-concatenate-reverse (cons str buffer))))
               (define prefix
                 (strftime (%current-logfile-date-format)
                           (localtime (current-time))))
 
-              ;; Make exactly one 'display' call per line to make sure we
-              ;; don't create several entries for each line.
-              (for-each (lambda (line)
-                          (display (string-append prefix line "\n")
-                                   (log-output-port)))
-                        lines)
+              ;; Note: We want to render as many newlinew as present in STR,
+              ;; so neither 'string-split' nor 'string-tokenize' helps.
+              (let loop ((str str))
+                (let* ((index (string-index str #\newline))
+                       (line  (if index (string-take str (+ 1 index)) str)))
+                  (unless (string-null? str)
+                    ;; Make exactly one 'display' call per line to make sure we
+                    ;; don't create several entries for each line.
+                    (display (string-append prefix line) (log-output-port))
+                    (when index
+                      (loop (string-drop str (+ index 1)))))))
+
               (set! buffer '())))))
 
     ;; Flush output.
