@@ -35,8 +35,6 @@
   #:use-module (shepherd comm)
   #:use-module (shepherd config)
   #:use-module (shepherd system)
-  #:replace (system
-             system*)
   #:export (<service>
             service?
             canonical-name
@@ -205,7 +203,10 @@ respawned, shows that it has been respawned more than TIMES in SECONDS."
   (stop-delay? #:init-keyword #:stop-delay?
 	       #:init-value #f)
   ;; The times of the last respawns, most recent first.
-  (last-respawns #:init-form '()))
+  (last-respawns #:init-form '())
+  ;; A replacement for when this service is stopped.
+  (replacement #:init-keyword #:replacement
+               #:init-value #f))
 
 (define (service? obj)
   "Return true if OBJ is a service."
@@ -288,24 +289,24 @@ wire."
 ;; Enable the service, allow it to get started.
 (define-method (enable (obj <service>))
   (slot-set! obj 'enabled? #t)
-  (local-output "Enabled service ~a." (canonical-name obj)))
+  (local-output (l10n "Enabled service ~a.") (canonical-name obj)))
 
 ;; Disable the service, make it unstartable.
 (define-method (disable (obj <service>))
   (slot-set! obj 'enabled? #f)
-  (local-output "Disabled service ~a." (canonical-name obj)))
+  (local-output (l10n "Disabled service ~a.") (canonical-name obj)))
 
 ;; Start the service, including dependencies.
 (define-method (start (obj <service>) . args)
   (cond ((running? obj)
-	 (local-output "Service ~a is already running."
+	 (local-output (l10n "Service ~a is already running.")
 		       (canonical-name obj)))
 	((not (enabled? obj))
-	 (local-output "Service ~a is currently disabled."
+	 (local-output (l10n "Service ~a is currently disabled.")
 		       (canonical-name obj)))
 	((let ((conflicts (conflicts-with-running obj)))
 	   (or (null? conflicts)
-	       (local-output "Service ~a conflicts with running services ~a."
+	       (local-output (l10n "Service ~a conflicts with running services ~a.")
 			     (canonical-name obj)
 			     (map canonical-name conflicts)))
 	   (not (null? conflicts)))
@@ -317,7 +318,7 @@ wire."
 		;; Resolve all dependencies.
 		(find (negate start) (required-by obj))))
 	   (if problem
-	       (local-output "Service ~a depends on ~a."
+	       (local-output (l10n "Service ~a depends on ~a.")
 			     (canonical-name obj)
 			     problem)
                (call-with-blocked-asyncs
@@ -341,56 +342,89 @@ wire."
 			 (canonical-name obj)))))
   (slot-ref obj 'running))
 
+(define (replace-service old-service new-service)
+  "Replace OLD-SERVICE with NEW-SERVICE in the services registry.  This
+completely removes all references to OLD-SERVICE before registering
+NEW-SERVICE."
+  (define (remove-service name)
+    (let* ((old (hashq-ref %services name))
+           (new (delete old-service old)))
+      (if (null? new)
+          (hashq-remove! %services name)
+          (hashq-set! %services name new))))
+  (when new-service
+    (for-each remove-service (provided-by old-service))
+    (register-services new-service)))
+
+(define (required-by? service dependent)
+  "Returns #t if DEPENDENT directly requires SERVICE in order to run.  Returns
+#f otherwise."
+  (and (find (lambda (dependency)
+               (memq dependency (provided-by service)))
+             (required-by dependent))
+       #t))
+
 ;; Stop the service, including services that depend on it.  If the
 ;; latter fails, continue anyway.  Return `#f' if it could be stopped.
-(define-method (stop (obj <service>) . args)
+(define-method (stop (service <service>) . args)
+  "Stop SERVICE, and any services which depend on it.  Returns a list of
+canonical names for all of the services which have been stopped (including
+transitive dependent services).  This method will print a warning if SERVICE
+is not already running, and will return SERVICE's canonical name in a list."
   ;; Block asyncs so the SIGCHLD handler doesn't execute concurrently.
-  ;; Notably, that makes sure the handler processes the SIGCHLD for OBJ's
-  ;; process once we're done; otherwise, it could end up respawning OBJ.
+  ;; Notably, that makes sure the handler processes the SIGCHLD for SERVICE's
+  ;; process once we're done; otherwise, it could end up respawning SERVICE.
   (call-with-blocked-asyncs
    (lambda ()
-     (if (not (running? obj))
-         (local-output "Service ~a is not running." (canonical-name obj))
-         (if (slot-ref obj 'stop-delay?)
+     (if (not (running? service))
+         (begin
+           (local-output (l10n "Service ~a is not running.")
+                         (canonical-name service))
+           (list (canonical-name service)))
+         (if (slot-ref service 'stop-delay?)
              (begin
-               (slot-set! obj 'waiting-for-termination? #t)
-               (local-output "Service ~a pending to be stopped."
-                             (canonical-name obj)))
-             (begin
-               ;; Stop services that depend on it.
-               (for-each-service
-                (lambda (serv)
-                  (and (running? serv)
-                       (for-each (lambda (sym)
-                                   (and (memq sym (provided-by obj))
-                                        (stop serv)))
-                                 (required-by serv)))))
-
+               (slot-set! service 'waiting-for-termination? #t)
+               (local-output (l10n "Service ~a pending to be stopped.")
+                             (canonical-name service))
+               (list (canonical-name service)))
+             (let ((name (canonical-name service))
+                   (stopped-dependents (fold-services (lambda (other acc)
+                                                        (if (and (running? other)
+                                                                 (required-by? service other))
+                                                            (append (stop other) acc)
+                                                            acc))
+                                                      '())))
                ;; Stop the service itself.
                (catch #t
                  (lambda ()
-                   (apply (slot-ref obj 'stop)
-                          (slot-ref obj 'running)
+                   (apply (slot-ref service 'stop)
+                          (slot-ref service 'running)
                           args))
                  (lambda (key . args)
                    ;; Special case: 'root' may quit.
-                   (and (eq? root-service obj)
+                   (and (eq? root-service service)
                         (eq? key 'quit)
                         (apply quit args))
                    (caught-error key args)))
 
-               ;; OBJ is no longer running.
-               (slot-set! obj 'running #f)
+               ;; SERVICE is no longer running.
+               (slot-set! service 'running #f)
 
                ;; Reset the list of respawns.
-               (slot-set! obj 'last-respawns '())
+               (slot-set! service 'last-respawns '())
+
+               ;; Replace the service with its replacement, if it has one
+               (let ((replacement (slot-ref service 'replacement)))
+                 (when replacement
+                   (replace-service service replacement)))
 
                ;; Status message.
-               (let ((name (canonical-name obj)))
-                 (if (running? obj)
-                     (local-output "Service ~a could not be stopped." name)
-                     (local-output "Service ~a has been stopped." name))))))
-     (slot-ref obj 'running))))
+               (if (running? service)
+                   (local-output (l10n "Service ~a could not be stopped.")
+                                 name)
+                   (local-output (l10n "Service ~a has been stopped.")
+                                 name))
+               (cons name stopped-dependents)))))))
 
 ;; Call action THE-ACTION with ARGS.
 (define-method (action (obj <service>) the-action . args)
@@ -401,10 +435,9 @@ wire."
       ;; Restarting is done in the obvious way.
       ((restart)
        (lambda (running . args)
-         (if running
-             (stop obj)
-             (local-output "~a was not running." (canonical-name obj)))
-         (apply start obj args)))
+         (let ((stopped-services (stop obj)))
+           (for-each start stopped-services)
+           #t)))
       ((status)
        ;; Return the service itself.  It is automatically converted to an sexp
        ;; via 'result->sexp' and sent to the client.
@@ -428,20 +461,11 @@ wire."
   (let ((proc (or (and=> (lookup-action obj the-action)
                          action-procedure)
 		  default-action)))
-    ;; Calling default-action will be allowed even when the service is
-    ;; not running, as it provides generally useful functionality and
-    ;; information.
-    ;; FIXME: Why should the user-implementations not be allowed to be
-    ;; called this way?
+    ;; Invoking THE-ACTION is allowed even when the service is not running, as
+    ;; it provides generally useful functionality and information.
     (catch #t
       (lambda ()
-        (cond ((eq? proc default-action)
-               (apply default-action obj args))
-              ((not (running? obj))
-               (local-output "Service ~a is not running." (canonical-name obj))
-               #f)
-              (else
-               (apply proc (slot-ref obj 'running) args))))
+        (apply proc (slot-ref obj 'running) args))
       (lambda (key . args)
         ;; Special case: 'root' may quit.
         (and (eq? root-service obj)
@@ -482,7 +506,7 @@ wire."
 		     (action-list obj)))
       (else
        ;; FIXME: Implement doc-help.
-       (local-output "Unknown keyword.  Try 'doc root help'.")))))
+       (local-output (l10n "Unknown keyword.  Try 'doc root help'."))))))
 
 ;; Return a list of services that conflict with OBJ.
 (define-method (conflicts-with (obj <service>))
@@ -590,13 +614,6 @@ results."
                (apply action service the-action args))
              which-services))))
 
-;; EINTR-safe versions of 'system' and 'system*'.
-
-(define system*
-  (EINTR-safe (@ (guile) system*)))
-
-(define system
-  (EINTR-safe (@ (guile) system)))
 
 
 
@@ -608,21 +625,23 @@ results."
 (define (handle-unknown . args)
   (let ((unknown (lookup-running 'unknown)))
     ;; FIXME: Display message if no unknown service.
-    (if unknown
-	(apply-to-args args
-	    (case-lambda
-	     ;; Start or stop.
-	     ((service-symbol start/stop args)
-	      (if (defines-action? unknown start/stop)
-		  (apply action unknown start/stop service-symbol args)
-		;; FIXME: Bad message.
-		(local-output "Cannot ~a ~a." start/stop service-symbol)))
-	     ;; Action.
-	     ((service-symbol action-sym the-action args)
-	      (assert (eq? action-sym 'action))
-	      (if (defines-action? unknown 'action)
-		  (apply action unknown 'action service-symbol the-action args)
-		(local-output "No service provides ~a." service-symbol))))))))
+    (when unknown
+      (apply (case-lambda
+	       ;; Start or stop.
+	       ((service-symbol start/stop args)
+	        (if (defines-action? unknown start/stop)
+		    (apply action unknown start/stop service-symbol args)
+		    ;; FIXME: Bad message.
+		    (local-output "Cannot ~a ~a." start/stop service-symbol)))
+	       ;; Action.
+	       ((service-symbol action-sym the-action args)
+	        (assert (eq? action-sym 'action))
+	        (if (defines-action? unknown 'action)
+		    (apply action unknown 'action service-symbol
+                           the-action args)
+		    (local-output (l10n "No service provides ~a.")
+                                  service-symbol))))
+             args))))
 
 ;; Check if any of SERVICES is running.  If this is the case, return
 ;; it.  If none, return `#f'.  Only the first one found will be
@@ -947,6 +966,16 @@ Return #f if service is not found."
           (eq? name (canonical-name service)))
         services))
 
+(define (fold-services proc init)
+  "Apply PROC to the registered services to build a result, and return that
+result.  Works in a manner akin to `fold' from SRFI-1."
+  (hash-fold (lambda (name services acc)
+               (let ((service (lookup-canonical-service name services)))
+                 (if service
+                     (proc service acc)
+                     acc)))
+             init %services))
+
 (define (for-each-service proc)
   "Call PROC for each registered service."
   (hash-for-each (lambda (name services)
@@ -981,21 +1010,19 @@ returned in unspecified."
   (hashq-ref %services name '()))
 
 (define waitpid*
-  (let ((waitpid (EINTR-safe waitpid)))
-    (lambda (what flags)
-      "Like 'waitpid', but EINTR-safe, and return (0 . _) when there's no
-child left."
-      (catch 'system-error
-        (lambda ()
-          (waitpid what flags))
-        (lambda args
-          ;; Did we get ECHILD or something?  If we did, that's a problem,
-          ;; because this procedure is supposed to be called only upon
-          ;; SIGCHLD.
-          (let ((errno (system-error-errno args)))
-            (local-output "warning: 'waitpid' ~a failed unexpectedly: ~a"
-                          what (strerror errno))
-            '(0 . #f)))))))
+  (lambda (what flags)
+    "Like 'waitpid', and return (0 . _) when there's no child left."
+    (catch 'system-error
+      (lambda ()
+        (waitpid what flags))
+      (lambda args
+        ;; Did we get ECHILD or something?  If we did, that's a problem,
+        ;; because this procedure is supposed to be called only upon
+        ;; SIGCHLD.
+        (let ((errno (system-error-errno args)))
+          (local-output (l10n "warning: 'waitpid' ~a failed unexpectedly: ~a")
+                        what (strerror errno))
+          '(0 . #f))))))
 
 (define (handle-SIGCHLD signum)
   "Handle SIGCHLD, possibly by respawning the service that just died, or
@@ -1034,7 +1061,7 @@ then disable it."
       (if (not (slot-ref serv 'waiting-for-termination?))
           (begin
             ;; Everything is okay, start it.
-            (local-output "Respawning ~a."
+            (local-output (l10n "Respawning ~a.")
                           (canonical-name serv))
             (slot-set! serv 'last-respawns
                        (cons (current-time)
@@ -1044,37 +1071,53 @@ then disable it."
           ;; termination.  The `running' slot has already
           ;; been set to `#f' by `stop'.
           (begin
-            (local-output "Service ~a terminated."
+            (local-output (l10n "Service ~a terminated.")
                           (canonical-name serv))
             (slot-set! serv 'waiting-for-termination? #f)))
       (begin
-        (local-output "Service ~a has been disabled."
+        (local-output (l10n "Service ~a has been disabled.")
                       (canonical-name serv))
         (when (respawn? serv)
-          (local-output "  (Respawning too fast.)"))
+          (local-output (l10n "  (Respawning too fast.)")))
         (slot-set! serv 'enabled? #f))))
 
 ;; Add NEW-SERVICES to the list of known services.
 (define (register-services . new-services)
+  "Add NEW-SERVICES to the list of known services.  If a service has already
+been registered, arrange to have it replaced when it is next stopped.  If it
+is currently stopped, replace it immediately."
   (define (register-single-service new)
     ;; Sanity-checks first.
     (assert (list-of-symbols? (provided-by new)))
     (assert (list-of-symbols? (required-by new)))
     (assert (boolean? (respawn? new)))
-    ;; Canonical name actually must be canonical.  (FIXME: This test
-    ;; is incomplete, since we may add a service later that makes it
-    ;; non-cannonical.)
-    (assert (null? (lookup-services (canonical-name new))))
-    ;; FIXME: Verify consistency: Check that there are no circular
-    ;; dependencies, check for bogus conflicts/dependencies, whatever
-    ;; else makes sense.
 
-    ;; Insert into the hash table.
-    (for-each (lambda (name)
-		(let ((old (lookup-services name)))
-		  ;; Actually add the new service now.
-		  (hashq-set! %services name (cons new old))))
-	      (provided-by new)))
+    ;; FIXME: Just because we have a unique canonical name now doesn't mean it
+    ;; will remain unique as other services are added. Whenever a service is
+    ;; added it should check that it's not conflicting with any already
+    ;; registered canonical names.
+    (match (lookup-services (canonical-name new))
+      (() ;; empty, so we can safely add ourselves
+       (for-each (lambda (name)
+		   (let ((old (lookup-services name)))
+		     (hashq-set! %services name (cons new old))))
+	         (provided-by new)))
+      ((old) ;; one service registered, so it may be an old version of us
+       (cond
+        ((not (eq? (canonical-name new) (canonical-name old)))
+         (local-output
+          "Cannot register service ~a: canonical name is not unique."
+          (canonical-name new))
+         (throw 'non-canonical-name))
+        ((running? old)
+         (slot-set! old 'replacement new))
+        (#:else
+         (replace-service old new))))
+      (_ ;; in any other case, there are too many services to register
+       (local-output
+        "Cannot register service ~a: canonical name is not unique."
+        (canonical-name new))
+       (throw 'non-canonical-name))))
 
   (for-each register-single-service new-services))
 
@@ -1121,10 +1164,10 @@ requested to be removed."
     (cond ((eq? name 'all)
            ;; Special 'remove all' case.
            (let ((pairs (service-pairs)))
-             (local-output "Unloading all optional services: '~a'..."
+             (local-output (l10n "Unloading all optional services: '~a'...")
                            (map car pairs))
              (for-each deregister (map cdr pairs))
-             (local-output "Done.")))
+             (local-output (l10n "Done."))))
           (else
            ;; Removing only one service.
            (match (lookup-services name)
@@ -1133,20 +1176,20 @@ requested to be removed."
              ((service)             ; only SERVICE provides NAME
               ;; Are we removing a user service…
               (if (eq? (canonical-name service) name)
-                  (local-output "Removing service '~a'..." name)
+                  (local-output (l10n "Removing service '~a'...") name)
                   ;; or a virtual service?
                   (local-output
                    "Removing service '~a' providing '~a'..."
                    (canonical-name service) name))
               (deregister service)
-              (local-output "Done."))
+              (local-output (l10n "Done.")))
              ((services ...)            ; ambiguous NAME
               (local-output
                "Not unloading: '~a' names several services: '~a'."
                name (map canonical-name services))))))))
 
 (define (load-config file-name)
-  (local-output "Loading ~a." file-name)
+  (local-output (l10n "Loading ~a.") file-name)
   ;; Every action is protected anyway, so no need for a `catch'
   ;; here.  FIXME: What about `quit'?
   (load-in-user-module file-name))
@@ -1195,7 +1238,8 @@ where prctl/PR_SET_CHILD_SUBREAPER is unsupported."
                       (let ((running (slot-ref service 'running)))
                         (when (and (integer? running)
                                    (not (process-exists? running)))
-                            (local-output "PID ~a (~a) is dead!" running (canonical-name service))
+                          (local-output (l10n "PID ~a (~a) is dead!")
+                                        running (canonical-name service))
                             (respawn-service service))))))
 
 (define root-service
@@ -1209,7 +1253,7 @@ where prctl/PR_SET_CHILD_SUBREAPER is unsupported."
                 (display-version))
 	      #t)
     #:stop (lambda (unused . args)
-	     (local-output "Exiting shepherd...")
+	     (local-output (l10n "Exiting shepherd..."))
 	     ;; Prevent that we try to stop ourself again.
 	     (slot-set! root-service 'running #f)
              (shutdown-services)
@@ -1247,7 +1291,7 @@ Clients such as 'herd' can read it and format it in a human-readable way."
         (catch 'quit
           (cut stop root-service)
           (lambda (key)
-            (local-output "Halting...")
+            (local-output (l10n "Halting..."))
             (halt)))))
      ;; Power off.
      (power-off
@@ -1256,7 +1300,7 @@ Clients such as 'herd' can read it and format it in a human-readable way."
         (catch 'quit
           (cut stop root-service)
           (lambda (key)
-            (local-output "Shutting down...")
+            (local-output (l10n "Shutting down..."))
             (power-off)))))
      ;; Evaluate arbitrary code.
      (load
@@ -1269,7 +1313,7 @@ dangerous.  You have been warned."
 potentially dangerous, be careful."
       (lambda (running str)
         (let ((exp (call-with-input-string str read)))
-          (local-output "Evaluating user expression ~a."
+          (local-output (l10n "Evaluating user expression ~a.")
                         (call-with-output-string
                           (lambda (port)
                             (truncated-print exp port #:width 50))))
@@ -1284,7 +1328,7 @@ removing them if needed."
         (deregister-service service-name)))
      (reload
       "Unload all services, then load from FILE-NAME into shepherd.  This
-is potentialy dangerous.  You have been warned."
+is potentially dangerous.  You have been warned."
       (lambda (running file-name)
         (and (deregister-service "all") ; unload all services
              (load-config file-name)))) ; reload from FILE-NAME
@@ -1292,13 +1336,13 @@ is potentialy dangerous.  You have been warned."
      (daemonize
       "Go into the background.  Be careful, this means that a new
 process will be created, so shepherd will not get SIGCHLD signals anymore
-if previously spawned childs terminate.  Therefore, this action should
-usually only be used (if at all) *before* childs get spawned for which
+if previously spawned children terminate.  Therefore, this action should
+usually only be used (if at all) *before* children get spawned for which
 we want to receive these signals."
       (lambda (running)
         (case (getpid)
           ((1)
-           (local-output "Running as PID 1, so not daemonizing."))
+           (local-output (l10n "Running as PID 1, so not daemonizing.")))
           (else
            (if (zero? (primitive-fork))
                (begin
@@ -1329,6 +1373,6 @@ when in interactive mode, i.e. with `--socket=none'."
      (restart
       "This does not work for the 'root' service."
       (lambda (running)
-	(local-output "You must be kidding."))))))
+	(local-output (l10n "You must be kidding.")))))))
 
 (register-services root-service)
