@@ -23,14 +23,15 @@
 (define-module (shepherd support)
   #:use-module (shepherd config)
   #:use-module (ice-9 match)
-  #:export (call/ec
+  #:export (buffering-mode
+
+            call/ec
             caught-error
             assert
             label
 
             catch-system-error
             with-system-error-handling
-            EINTR-safe
             with-atomic-file-output
             mkdir-p
             with-directory-excursion
@@ -62,6 +63,23 @@
 
             verify-dir))
 
+(define-syntax-rule (if-2.0 subsequent alternate)
+  "Expand to SUBSEQUENT when using Guile 2.0, and to ALTERNATE otherwise."
+  (cond-expand
+    ((and guile-2 (not guile-2.2)) subsequent)
+    (else alternate)))
+
+(define-syntax buffering-mode
+  (syntax-rules (line block none)
+    "Return the appropriate buffering mode depending on whether we're on Guile
+2.0 or later."
+    ((_ line)
+     (if-2.0 _IOLBF 'line))
+    ((_ block)
+     (if-2.0 _IOFBF 'block))
+    ((_ none)
+     (if-2.0 _IONBF 'none))))
+
 ;; Implement `call-with-escape-continuation' with `catch' and `throw'.
 ;; FIXME: Multiple return values.
 (define (call/ec proc)
@@ -93,7 +111,7 @@
 (define-syntax-rule (assert EXPR)
   (and (not EXPR)
        (begin
-	 (local-output "Assertion ~a failed." 'EXPR)
+	 (local-output (l10n "Assertion ~a failed.") 'EXPR)
 	 (throw 'assertion-failed))))
 
 ;; Recursive procedures.
@@ -127,19 +145,6 @@ turned into user error messages."
    (lambda ()
      body ...)))
 
-(define (EINTR-safe proc)
-  "Wrap PROC so that if a 'system-error' exception with EINTR is raised (that
-was possible up to Guile 2.0.9 included) the call to PROC is restarted."
-  (lambda args
-    (let loop ()
-      (catch 'system-error
-        (lambda ()
-          (apply proc args))
-        (lambda args
-          (if (= EINTR (system-error-errno args))
-              (loop)
-              (apply throw args)))))))
-
 (define (with-atomic-file-output file proc)       ;copied from Guix
   "Call PROC with an output port for the file that is going to replace FILE.
 Upon success, FILE is atomically replaced by what has been written to the
@@ -214,14 +219,13 @@ output port, and PROC's result is returned."
     (debug-disable 'warn-deprecated))
 
   ;; In Guile 2.2+, the locale is installed by default.
-  (cond-expand
-    (guile-2.0 (false-if-exception (setlocale LC_ALL "")))
-    (else      #t))
+  (if-2.0 (false-if-exception (setlocale LC_ALL ""))
+          #t)
 
   (bindtextdomain %gettext-domain %localedir)
   (textdomain %gettext-domain)
-  (setvbuf (current-output-port) _IOLBF)
-  (setvbuf (current-error-port) _IOLBF))
+  (setvbuf (current-output-port) (buffering-mode line))
+  (setvbuf (current-error-port) (buffering-mode line)))
 
 ;; Localized version of STR.
 (define l10n gettext)
@@ -229,7 +233,7 @@ output port, and PROC's result is returned."
 ;; Display some text and a newline.
 (define-syntax-rule (local-output format-string args ...)
   (begin
-    (format #t (gettext format-string) args ...)
+    (format #t format-string args ...)
     (newline)))
 
 (define* (display-version #:optional (program-name (program-name)))
@@ -396,5 +400,5 @@ directory are not checked."
     (let ((dir-stat (stat dir)))
       (unless (and (= (stat:uid dir-stat) (getuid))
                    (= (stat:perms dir-stat) #o700))
-        (local-output "Socket directory setup is insecure.")
+        (local-output (l10n "Socket directory setup is insecure."))
         (exit 1)))))
